import math
from abc import ABC


class DiscMorrisonConstants(ABC):
    # pylint: disable = invalid-name, too-many-instance-attributes
    GRAVITY = -9.81
    MASS = 0.175  # in kg
    RHO = 1.23
    A = 0.0568
    CL0 = 0.15
    CLalpha = 1.4
    CD0 = 0.08
    CDalpha = 2.72
    ALPHA0 = -4

    @classmethod
    def CL(cls, alpha: float) -> float:
        return cls.CL0 + cls.CLalpha * math.radians(alpha)

    @classmethod
    def CD(cls, alpha: float) -> float:
        return cls.CD0 + cls.CDalpha * math.pow(math.radians(alpha - cls.ALPHA0), 2)


class DiscMorrisonUltrastar(DiscMorrisonConstants):
    # pylint: disable = invalid-name
    RADIUS = 0.14  # in m
    MASS = 0.175  # in kg
    # A = RADIUS**2 * math.pi


class DiscMorrisonUltrastarCode(DiscMorrisonUltrastar):
    # pylint: disable = invalid-name
    # The paper do use a different CL0 value in the code then in the showed calculations.
    CL0 = 0.1



from copy import deepcopy
from typing import Any, List, Sequence, Tuple

from flying_discs.disc_position import DiscPosition
from flying_discs.morrison.disc_morrison_linear import DiscMorrisonLinear
from flying_discs.morrison.morrison_constants import DiscMorrisonConstants
from flying_discs.utils import quadratic_bezier, rotate_points_around_mid_point


class DiscMorrisonBezier(DiscMorrisonLinear):
    def __init__(self, constants: DiscMorrisonConstants, x: float = 0.0, y: float = 0.0, z: float = 1.7) -> None:
        # pylint : disable=too-many-instance-attributes
        super().__init__(constants, x, y, z)
        self._bezier_points: List[Tuple[float, float]] = []
        self._linear_trajectory: Sequence[DiscPosition] = []
        self._rotated_points: List[Tuple[float, float]] = []
        self.rotation_angle: float = 0.0
        self.factor: float = 0.0

    @staticmethod
    def _determine_bezier_points(
        trajectory: Sequence[DiscPosition],
        rotated_points: List[Tuple[float, float]],
        tmax: int,
        factor: float,
    ) -> List[Tuple[float, float]]:
        x0 = trajectory[0].x
        y0 = trajectory[0].y
        idx = int(tmax * factor)
        x1 = rotated_points[idx][0]
        y1 = rotated_points[idx][1]
        x2 = trajectory[-1].x
        y2 = trajectory[-1].y
        bezier_points = [(x0, y0), (x1, y1), (x2, y2)]
        return bezier_points

    def _calculate_trajectory(self, v0: float, alpha: float, deltaT: float, **kwargs: Any) -> Sequence[DiscPosition]:
        trajectory = super()._calculate_trajectory(v0, alpha, deltaT, angle=kwargs["angle"])
        self._linear_trajectory = deepcopy(trajectory)
        rotated_points = rotate_points_around_mid_point(
            [(p.x, p.y) for p in trajectory],
            kwargs["rotation_angle"],
        )
        self._rotated_points = rotated_points

        tmax = len(trajectory) - 1
        self._bezier_points = self._determine_bezier_points(
            trajectory,
            rotated_points,
            tmax,
            kwargs["factor"],
        )
        for t in range(tmax):
            x, y = quadratic_bezier(
                t / tmax,
                self._bezier_points[0],
                self._bezier_points[1],
                self._bezier_points[2],
            )
            trajectory[t].x = x
            trajectory[t].y = y

        return trajectory

    def calculate_trajectory(self, timescale: float, **kwargs: Any) -> Sequence[DiscPosition]:
        self.throw_direction = kwargs["direction"]
        self.throw_aoa = kwargs["alpha"]
        self.throw_power = kwargs["power"]
        self.rotation_angle = kwargs["rotation_angle"]
        self.factor = kwargs["factor"]
        self.trajectory = self._calculate_trajectory(
            self.throw_power,
            self.throw_aoa,
            timescale,
            angle=self.throw_direction,
            rotation_angle=self.rotation_angle,
            factor=self.factor,
        )
        return self.trajectory

    def calculate_trajectory_to_position(
        self, x: float, y: float, timescale: float, **kwargs: Any
    ) -> Sequence[DiscPosition]:
        raise NotImplementedError()



import math
from unittest.mock import MagicMock

import pytest

from flying_discs.morrison.disc_morrison_linear import DiscMorrisonLinear
from flying_discs.morrison.morrison_constants import DiscMorrisonUltrastar


@pytest.fixture
def setup() -> DiscMorrisonLinear:
    return DiscMorrisonLinear(DiscMorrisonUltrastar())


def test_calculate_trajectory(setup: DiscMorrisonLinear) -> None:
    disc = setup

    disc._calculate_trajectory = MagicMock(spec_set=disc._calculate_trajectory)

    disc.calculate_trajectory(0.033, direction=10.0, alpha=2.5, v0=20.0)

    assert disc.throw_direction == 10.0
    assert disc.throw_aoa == 2.5
    assert disc.throw_power == 20.0

    disc._calculate_trajectory.assert_called_once_with(20.0, 2.5, 0.033, angle=10.0)


def test_calculate_trajectory_to_position(setup: DiscMorrisonLinear) -> None:
    disc = setup

    disc._approximate_trajectory_to_target_x_y = MagicMock(
        spec_set=disc._approximate_trajectory_to_target_x_y, return_value=([], 10.0, 2.5)
    )

    disc.calculate_trajectory_to_position(1, 1, 0.033)

    assert disc.throw_direction == math.radians(45)
    assert disc.throw_aoa == 2.5
    assert disc.throw_power == 10.0
    assert disc.trajectory == []

    disc._approximate_trajectory_to_target_x_y.assert_called_once_with(1, 1, 0.033, angle=math.radians(45))


def test__calculate_trajectory() -> None:
    # TODO: Add test
    pass


def test__approximate_trajectory_to_target_x_y() -> None:
    # TODO: Add test
    pass



from flying_discs.morrison.disc_morrison_linear import DiscMorrisonLinear
from flying_discs.morrison.morrison_constants import DiscMorrisonUltrastar

x0 = 0
y0 = 0
z0 = 1

disc = DiscMorrisonLinear(DiscMorrisonUltrastar(), x0, y0, z0)

timescale = 0.033
angle_of_attack = 5
v0 = 14
direction = 0

trajectory = disc.calculate_trajectory(
    timescale,
    alpha=angle_of_attack,
    v0=v0,
    direction=direction,
)



import math

import pytest
from numpy.testing import assert_almost_equal

from flying_discs.utils import (
    angle_between_vectors,
    distance_v1_v2,
    quadratic_bezier,
    rotate_points_around_index,
    rotate_points_around_mid_point,
    rotate_points_around_origin
)


def test_angle_between_vectors() -> None:
    assert angle_between_vectors((10, 20), (20, 10)) == -angle_between_vectors((20, 10), (10, 20))
    assert angle_between_vectors((1, 0), (0, 1)) == math.radians(90)
    assert angle_between_vectors((0, 1), (1, 0)) == math.radians(-90)
    assert angle_between_vectors((0, 1), (0, -1)) == math.radians(180)
    assert angle_between_vectors((1, 0), (-1, 0)) == math.radians(180)


def test_distance_v1_v2() -> None:
    assert distance_v1_v2(*(10, 10), *(10, 10)) == 0
    assert distance_v1_v2(*(10, 10), *(0, 0)) == pytest.approx(14.142136)
    assert distance_v1_v2(*(10, 10), *(-10, -10)) == pytest.approx(28.284271)


def test_quadratic_bezier() -> None:
    p0 = (0.0, 0.0)
    p1 = (0.0, 2.0)
    p2 = (2.0, 2.0)

    # dt has to be in between [0,1]
    result = quadratic_bezier(0, p0, p1, p2)

    assert result == p0

    result = quadratic_bezier(1, p0, p1, p2)

    assert result == p2

    result = quadratic_bezier(0.5, p0, p1, p2)

    assert result == (0.5, 1.5)


def test_round_points_around_mid_origin() -> None:
    testee = [(1.0, 1.0), (2.0, 2.0), (3.0, 3.0)]
    result = rotate_points_around_origin(testee, 180)
    expected = [(-1.0, -1.0), (-2.0, -2.0), (-3.0, -3.0)]

    assert_almost_equal(result, expected)

    testee = [(1.0, 1.0), (2.0, 2.0), (3.0, 3.0)]
    result = rotate_points_around_origin(testee, 90)
    expected = [(-1.0, 1.0), (-2.0, 2.0), (-3.0, 3.0)]

    assert_almost_equal(result, expected)

    testee = [(1.0, 1.0), (2.0, 2.0), (3.0, 3.0)]
    result = rotate_points_around_origin(testee, 270)
    expected = [(1.0, -1.0), (2.0, -2.0), (3.0, -3.0)]

    assert_almost_equal(result, expected)

    testee = [(1.0, 1.0), (2.0, 2.0), (3.0, 3.0)]
    result = rotate_points_around_origin(testee, 360)

    assert_almost_equal(result, testee)

    testee = [(1.0, 1.0), (2.0, 2.0), (3.0, 3.0)]
    result = rotate_points_around_origin(testee, 45)
    expected = [(0.0, 1.4142), (0.0, 2.8284), (0.0, 4.2426)]

    assert_almost_equal(result, expected, decimal=4)


def test_rotate_points_around_index() -> None:
    testee = [(1.0, 1.0), (2.0, 2.0), (3.0, 3.0)]
    result = rotate_points_around_index(testee, 180, 0)
    expected = [(1.0, 1.0), (0.0, 0.0), (-1.0, -1.0)]

    assert_almost_equal(result, expected)

    testee = [(1.0, 1.0), (2.0, 2.0), (3.0, 3.0)]
    result = rotate_points_around_index(testee, 180, -1)
    expected = [(5.0, 5.0), (4.0, 4.0), (3.0, 3.0)]

    assert_almost_equal(result, expected)

    testee = [(1.0, 1.0), (2.0, 2.0), (3.0, 3.0)]
    result = rotate_points_around_index(testee, 90, 0)
    expected = [(1.0, 1.0), (0.0, 2.0), (-1.0, 3.0)]

    assert_almost_equal(result, expected)

    testee = [(1.0, 1.0), (2.0, 2.0), (3.0, 3.0)]
    result = rotate_points_around_index(testee, 90, -1)
    expected = [(5.0, 1.0), (4.0, 2.0), (3.0, 3.0)]

    assert_almost_equal(result, expected)

    testee = [(1.0, 1.0), (2.0, 2.0), (3.0, 3.0)]
    result = rotate_points_around_index(testee, 270, 0)
    expected = [(1.0, 1.0), (2.0, 0.0), (3.0, -1.0)]

    assert_almost_equal(result, expected)

    testee = [(1.0, 1.0), (2.0, 2.0), (3.0, 3.0)]
    result = rotate_points_around_index(testee, 270, -1)
    expected = [(1.0, 5.0), (2.0, 4.0), (3.0, 3.0)]

    assert_almost_equal(result, expected)

    testee = [(1.0, 1.0), (2.0, 2.0), (3.0, 3.0)]
    result = rotate_points_around_index(testee, 360, 0)

    assert_almost_equal(result, testee)

    testee = [(1.0, 1.0), (2.0, 2.0), (3.0, 3.0)]
    result = rotate_points_around_index(testee, 360, -1)

    assert_almost_equal(result, testee)

    testee = [(1.0, 1.0), (2.0, 2.0), (3.0, 3.0)]
    result = rotate_points_around_index(testee, 45, 0)
    expected = [(1.0, 1.0), (1.0, 2.4142), (1.0, 3.8284)]

    assert_almost_equal(result, expected, decimal=4)

    testee = [(1.0, 1.0), (2.0, 2.0), (3.0, 3.0)]
    result = rotate_points_around_index(testee, 45, -1)
    expected = [(3.0, 0.1716), (3.0, 1.5858), (3.0, 3.0)]

    assert_almost_equal(result, expected, decimal=4)


def test_round_points_around_mid_point() -> None:
    testee = [(1.0, 1.0), (2.0, 2.0), (3.0, 3.0)]
    result = rotate_points_around_mid_point(testee, 180)
    expected = [(3.0, 3.0), (2.0, 2.0), (1.0, 1.0)]

    assert_almost_equal(result, expected)

    testee = [(1.0, 1.0), (2.0, 2.0), (3.0, 3.0)]
    result = rotate_points_around_mid_point(testee, 90)
    expected = [(3.0, 1.0), (2.0, 2.0), (1.0, 3.0)]

    assert_almost_equal(result, expected)

    testee = [(1.0, 1.0), (2.0, 2.0), (3.0, 3.0)]
    result = rotate_points_around_mid_point(testee, 270)
    expected = [(1.0, 3.0), (2.0, 2.0), (3.0, 1.0)]

    assert_almost_equal(result, expected)

    testee = [(1.0, 1.0), (2.0, 2.0), (3.0, 3.0)]
    result = rotate_points_around_mid_point(testee, 360)

    assert_almost_equal(result, testee)

    testee = [(1.0, 1.0), (2.0, 2.0), (3.0, 3.0)]
    result = rotate_points_around_mid_point(testee, 45)
    expected = [(2.0, 0.5857), (2.0, 2.0), (2.0, 3.4142)]

    assert_almost_equal(result, expected, decimal=4)



import math
from typing import Any, Optional, Sequence

from flying_discs.disc import Disc
from flying_discs.disc_position import DiscPosition
from flying_discs.morrison.morrison_constants import DiscMorrisonConstants


class DiscMorrison(Disc):
    # pylint: disable = invalid-name,too-many-instance-attributes,too-many-locals
    def __init__(self, constants: DiscMorrisonConstants, x: float = 0.0, y: float = 0.0, z: float = 1.7) -> None:
        super().__init__(DiscPosition(x=x, y=y, z=z))
        self.constants = constants
        self.throw_direction = 0.0
        self.throw_power = 0.0
        self.throw_aoa = 0.0
        self._aoa = 2.5

    def _calculate_trajectory_step(
        self, deltaT: float, d: float, z: float, vd: float, vz: float, CD: float, CL: float
    ) -> DiscPosition:
        ad = -self.constants.RHO * math.pow(vd, 2) * self.constants.A * CD * deltaT
        az = (
            self.constants.RHO * math.pow(vd, 2) * self.constants.A * CL / 2 / self.constants.MASS
            + self.constants.GRAVITY
        ) * deltaT
        vd = vd + ad
        vz = vz + az
        d = d + vd * deltaT
        z = z + vz * deltaT

        return DiscPosition(0.0, d, z, 0.0, vd, vz, 0.0, ad, az)

    def _calculate_trajectory(
        self, v0: float, alpha: float, deltaT: float, **_: Optional[Any]
    ) -> Sequence[DiscPosition]:
        # In code of Morrision, only 2D is calculated, distance x and height y
        # In morrison y = z in here
        # In morrison x = d in here
        # d = distance
        vd = v0
        vz = 0.0
        d = 0.0
        z = self.z
        # We simulate approx. the function for x, y with a linear function towards the target
        CL = self.constants.CL(alpha=alpha)
        CD = self.constants.CD(alpha=alpha)
        trajectory = [DiscPosition(0.0, d, z, 0.0, vd, vz, 0.0, 0.0, 0.0)]
        while trajectory[-1].z > 0:
            current_step = trajectory[-1]
            next_position = self._calculate_trajectory_step(
                deltaT, current_step.y, current_step.z, current_step.vy, current_step.vz, CD, CL
            )
            trajectory.append(next_position)
        return trajectory

    def calculate_trajectory(self, timescale: float, **kwargs: Any) -> Sequence[DiscPosition]:
        self.throw_aoa = kwargs["alpha"]
        self.throw_power = kwargs["power"]
        self.trajectory = self._calculate_trajectory(self.throw_power, self.throw_aoa, timescale)
        return self.trajectory

    def calculate_trajectory_to_position(
        self, x: float, y: float, timescale: float, **kwargs: Any
    ) -> Sequence[DiscPosition]:
        raise NotImplementedError("This function is not implemented.")



from __future__ import annotations

from dataclasses import dataclass
from typing import Any


@dataclass
class DiscPosition:
    # pylint: disable = invalid-name,too-many-instance-attributes
    x: float = 0.0
    y: float = 0.0
    z: float = 0.0
    vx: float = 0.0
    vy: float = 0.0
    vz: float = 0.0
    ax: float = 0.0
    ay: float = 0.0
    az: float = 0.0

    def ground(self) -> None:
        self.z = 0.0
        self.vx = 0.0
        self.vy = 0.0
        self.vz = 0.0
        self.ax = 0.0
        self.ay = 0.0
        self.az = 0.0

    def reset(self) -> None:
        self.ground()
        self.x = 0.0
        self.y = 0.0

    @classmethod
    def position(cls, **kwargs: Any) -> DiscPosition:
        return cls(**kwargs)



import math
from typing import Any, Sequence, Tuple

import numpy as np

from flying_discs.disc_position import DiscPosition
from flying_discs.morrison.disc_morrison_base import DiscMorrison
from flying_discs.morrison.morrison_constants import DiscMorrisonConstants
from flying_discs.utils import angle_between_vectors, distance_v1_v2


class DiscMorrisonLinear(DiscMorrison):
    # pylint: disable = invalid-name,too-many-instance-attributes,too-many-locals
    def __init__(self, constants: DiscMorrisonConstants, x: float = 0.0, y: float = 0.0, z: float = 1.7) -> None:
        super().__init__(constants, x, y, z)
        self.constants = constants
        self.throw_direction = 0.0
        self.throw_power = 0.0
        self.throw_aoa = 0.0
        self._aoa = 2.5

    def _calculate_trajectory(self, v0: float, alpha: float, deltaT: float, **kwargs: Any) -> Sequence[DiscPosition]:
        trajectory = super()._calculate_trajectory(v0, alpha, deltaT)
        angle = kwargs["angle"]
        point: DiscPosition
        for i, point in enumerate(trajectory):
            if i == 0:
                new_vx = point.vy * math.cos(angle)
                new_vy = point.vy * math.sin(angle)
                point.vx = new_vx
                point.vy = new_vy
                point.x = self.x
                point.y = self.y
                continue
            new_ax = point.ay * math.cos(angle)
            new_ay = point.ay * math.sin(angle)
            point.ax = new_ax
            point.ay = new_ay
            point.vx = trajectory[i - 1].vx + point.ax
            point.vy = trajectory[i - 1].vy + point.ay
            point.x = trajectory[i - 1].x + point.vx * deltaT
            point.y = trajectory[i - 1].y + point.vy * deltaT
        return trajectory

    def calculate_trajectory(self, timescale: float, **kwargs: Any) -> Sequence[DiscPosition]:
        self.throw_direction = kwargs["direction"]
        self.throw_aoa = kwargs["alpha"]
        self.throw_power = kwargs["v0"]
        self.trajectory = self._calculate_trajectory(
            self.throw_power, self.throw_aoa, timescale, angle=self.throw_direction
        )
        return self.trajectory

    def calculate_trajectory_to_position(
        self, x: float, y: float, timescale: float, **kwargs: Any
    ) -> Sequence[DiscPosition]:
        self.throw_direction = angle_between_vectors((1, 0), (x - self.x, y - self.y))
        (
            self.trajectory,
            self.throw_power,
            self.throw_aoa,
        ) = self._approximate_trajectory_to_target_x_y(x, y, timescale, angle=self.throw_direction, **kwargs)
        return self.trajectory

    def _approximate_trajectory_to_target_x_y(
        self, x: float, y: float, timescale: float, **kwargs: Any
    ) -> Tuple[Sequence[DiscPosition], float, float]:
        dist = distance_v1_v2(x, y, self.x, self.y)
        v0 = 0.0
        distance_traveled = -math.inf
        height_at_target = -math.inf
        while distance_traveled < dist or height_at_target < 0:
            v0 += 0.1
            distance_traveled = -math.inf
            trajectory = self._calculate_trajectory(v0, self._aoa, timescale, **kwargs)
            distance_traveled = trajectory[-1].y
            d_ = np.array([p.y for p in trajectory])
            idx = np.where(d_ > dist)[0]
            if idx.size > 0:
                start_idx = idx[0]
                height_at_target = trajectory[start_idx].z
        return trajectory, v0, self._aoa



from dataclasses import dataclass

from flying_discs.disc_position import DiscPosition


@dataclass
class DiscMorrisonPosition(DiscPosition):
    # pylint: disable = invalid-name
    d: float = 0.0
    vd: float = 0.0
    ad: float = 0.0

    def ground(self) -> None:
        super().ground()
        self.d = 0.0
        self.vd = 0.0
        self.ad = 0.0



from abc import ABC, abstractmethod
from typing import Any, Sequence

from flying_discs.disc_position import DiscPosition
from flying_discs.disc_state import DiscState
from flying_discs.utils import distance_v1_v2


class Disc(ABC):
    # pylint: disable = invalid-name,too-many-instance-attributes,too-many-locals
    # TODO: manage radius, disc properties
    def __init__(self, position: DiscPosition, radius: float = 0.14) -> None:
        self.current_position = position
        self.radius = radius
        self.current_step = 0
        self.trajectory: Sequence[DiscPosition] = []
        self._state = DiscState.GROUNDED

    @abstractmethod
    def calculate_trajectory_to_position(
        self, x: float, y: float, timescale: float, **kwargs: Any
    ) -> Sequence[DiscPosition]:
        ...

    @abstractmethod
    def calculate_trajectory(self, timescale: float, **kwargs: Any) -> Sequence[DiscPosition]:
        ...

    @property
    def distance_to_target(self) -> float:
        return distance_v1_v2(self.x, self.target_x, self.y, self.target_y)

    @property
    def x(self) -> float:
        return self.current_position.x

    @property
    def y(self) -> float:
        return self.current_position.y

    @property
    def z(self) -> float:
        return self.current_position.z

    @property
    def vx(self) -> float:
        return self.current_position.vx

    @property
    def vy(self) -> float:
        return self.current_position.vy

    @property
    def vz(self) -> float:
        return self.current_position.vz

    @property
    def ax(self) -> float:
        return self.current_position.ax

    @property
    def ay(self) -> float:
        return self.current_position.ay

    @property
    def az(self) -> float:
        return self.current_position.az

    @property
    def state(self) -> DiscState:
        return self._state

    @property
    def target_x(self) -> float:
        if len(self.trajectory) > 0:
            return self.trajectory[-1].x
        # TODO: return none instead
        return 0.0

    @property
    def target_y(self) -> float:
        if len(self.trajectory) > 0:
            return self.trajectory[-1].y
        # TODO: return none instead
        return 0.0

    def set_position(self, **kwargs: Any) -> None:
        self.current_position = self.current_position.position(**kwargs)

    def move(self) -> None:
        if self._state == DiscState.AIRED:
            self.current_step += 1
            self.current_position = self.trajectory[self.current_step]
            if self.z <= 0.0:
                self.grounded()

    def release(self) -> None:
        self._state = DiscState.AIRED

    def grounded(self) -> None:
        self.current_position.ground()
        self._reset_disc_state()

    def reset(self) -> None:
        self.current_position.reset()
        self._reset_disc_state()

    def _reset_disc_state(self) -> None:
        self._state = DiscState.GROUNDED
        self.current_step = 0
        self.trajectory = []



from unittest.mock import MagicMock

import pytest

from flying_discs.morrison.disc_morrison_bezier import DiscMorrisonBezier
from flying_discs.morrison.morrison_constants import DiscMorrisonUltrastar


@pytest.fixture
def setup() -> DiscMorrisonBezier:
    return DiscMorrisonBezier(DiscMorrisonUltrastar())


def test_calculate_trajectory(setup: DiscMorrisonBezier) -> None:
    disc = setup

    disc._calculate_trajectory = MagicMock(spec_set=disc._calculate_trajectory)

    disc.calculate_trajectory(0.033, direction=10.0, alpha=2.5, power=10.0, rotation_angle=34, factor=0.5)

    assert disc.throw_direction == 10.0
    assert disc.throw_aoa == 2.5
    assert disc.throw_power == 10.0
    assert disc.rotation_angle == 34
    assert disc.factor == 0.5

    disc._calculate_trajectory.assert_called_once_with(10.0, 2.5, 0.033, angle=10.0, rotation_angle=34, factor=0.5)


def test__calculate_trajectory() -> None:
    # TODO: Add test
    pass


def test__determine_bezier_points() -> None:
    # TODO: Add test
    pass



from unittest.mock import MagicMock

import pytest

from flying_discs.disc_position import DiscPosition
from flying_discs.morrison.disc_morrison_base import DiscMorrison
from flying_discs.morrison.morrison_constants import DiscMorrisonUltrastar


@pytest.fixture
def setup() -> DiscMorrison:
    return DiscMorrison(DiscMorrisonUltrastar())


def test_calculate_trajectory_to_position(setup: DiscMorrison) -> None:
    disc = setup
    with pytest.raises(RuntimeError):
        disc.calculate_trajectory_to_position(10.0, 20.0, 0.33)


def test_calculate_trajectory(setup: DiscMorrison) -> None:
    disc = setup
    disc._calculate_trajectory = MagicMock(spec_set=disc._calculate_trajectory, return_value=[DiscPosition()])
    disc.calculate_trajectory(0.033, alpha=10.0, power=20.0)

    assert disc.throw_aoa == 10.0
    assert disc.throw_power == 20.0
    disc._calculate_trajectory.assert_called_once_with(20.0, 10.0, 0.033)
    assert disc.trajectory == [DiscPosition()]


def test__calculate_trajectory(setup: DiscMorrison) -> None:
    disc = setup
    disc._calculate_trajectory_step = MagicMock(
        spec_set=disc._calculate_trajectory_step,
        side_effect=[
            DiscPosition(z=1.0),
            DiscPosition(z=0.1),
            DiscPosition(z=0.0),
            DiscPosition(z=-1.0),
        ],
    )

    trajectory = disc._calculate_trajectory(20.0, 10.0, 0.033)

    assert trajectory[0] == DiscPosition(z=1.7, vy=20.0)
    assert trajectory[1] == DiscPosition(z=1.0)
    assert trajectory[2] == DiscPosition(z=0.1)
    assert trajectory[3] == DiscPosition(z=0.0)
    assert len(trajectory) == 4

    assert len(disc._calculate_trajectory_step.mock_calls) == 3



import pytest

from flying_discs.disc_position import DiscPosition


@pytest.fixture
def setup() -> DiscPosition:
    return DiscPosition(x=10.0, y=11.0, z=12.0, vx=1.0, vy=1.0, vz=1.1, ax=0.5, ay=0.5, az=0.5)


def test_ground(setup: DiscPosition) -> None:
    position = setup
    position.ground()

    assert position.x == 10.0
    assert position.y == 11.0
    assert position.z == 0.0
    assert position.vx == 0.0
    assert position.vy == 0.0
    assert position.vz == 0.0
    assert position.ax == 0.0
    assert position.ay == 0.0
    assert position.az == 0.0


def test_reset(setup: DiscPosition) -> None:
    position = setup
    position.reset()

    assert position.x == 0.0
    assert position.y == 0.0
    assert position.z == 0.0
    assert position.vx == 0.0
    assert position.vy == 0.0
    assert position.vz == 0.0
    assert position.ax == 0.0
    assert position.ay == 0.0
    assert position.az == 0.0


def test_position(setup: DiscPosition) -> None:
    position = setup
    position_kwargs = {
        "x": 10.0,
        "y": 11.0,
        "z": 12.0,
        "vx": 1.0,
        "vy": 1.0,
        "vz": 1.1,
        "ax": 0.5,
        "ay": 0.5,
        "az": 0.5,
    }
    new_position = position.position(**position_kwargs)
    assert id(new_position) != id(position)
    assert new_position == position



import pytest

from flying_discs.morrison.morrison_constants import DiscMorrisonUltrastar


def test_CL() -> None:
    # expected result were obtained by manual calculation
    assert 0.21108652381980153 == pytest.approx(DiscMorrisonUltrastar.CL(2.5))
    assert 0.17443460952792061 == pytest.approx(DiscMorrisonUltrastar.CL(1))
    assert 0.15 == pytest.approx(DiscMorrisonUltrastar.CL(0))
    assert 0.05226156188831754 == pytest.approx(DiscMorrisonUltrastar.CL(-4))


def test_CD() -> None:
    # expected result were obtained by manual calculation
    assert 0.11500663388188856 == pytest.approx(DiscMorrisonUltrastar.CD(2.5))
    assert 0.10071398454549618 == pytest.approx(DiscMorrisonUltrastar.CD(1))
    assert 0.09325695010911755 == pytest.approx(DiscMorrisonUltrastar.CD(0))
    assert 0.08 == pytest.approx(DiscMorrisonUltrastar.CD(-4))



from typing import Any
from unittest.mock import MagicMock

import pytest
from pytest_mock.plugin import MockerFixture

from flying_discs.disc import Disc
from flying_discs.disc_position import DiscPosition
from flying_discs.disc_state import DiscState


@pytest.fixture
def setup() -> Disc:
    class DiscTestImpl(Disc):
        def calculate_trajectory(self, timescale: float, **kwargs: Any) -> None:
            assert True

        def calculate_trajectory_to_position(self, x: float, y: float, timescale: float, **kwargs: Any) -> None:
            assert True

    return DiscTestImpl(DiscPosition(x=1.0, y=2.0, z=3.0), radius=0.28)


def test_distance_to_target(mocker: MockerFixture, setup: Disc) -> None:
    disc = setup
    disc.trajectory = [DiscPosition(x=2.0, y=3.0, z=4.0)]
    distance_v1_v2 = mocker.patch("flying_discs.disc.distance_v1_v2")

    disc.distance_to_target()
    distance_v1_v2.assert_called_once_with(disc.current_position.x, 2.0, disc.current_position.y, 3.0)


def test_position_properties(setup: Disc) -> None:
    disc = setup
    disc.current_position = DiscPosition(x=1.0, y=2.0, z=3.0, vx=4.0, vy=5.0, vz=6.0, ax=7.0, ay=8.0, az=9.0)

    assert disc.x == 1.0
    assert disc.y == 2.0
    assert disc.z == 3.0
    assert disc.vx == 4.0
    assert disc.vy == 5.0
    assert disc.vz == 6.0
    assert disc.ax == 7.0
    assert disc.ay == 8.0
    assert disc.az == 9.0


def test_traget_x_and_target_y(setup: Disc) -> None:
    disc = setup

    # Has no trajectory
    assert disc.target_x == 0.0
    assert disc.target_y == 0.0

    # Has trajectory
    trajectory = [
        DiscPosition(x=2.0, y=3.0, z=4.0),
        DiscPosition(x=3.0, y=4.0, z=0.0, ax=1.0),
    ]
    disc.trajectory = trajectory

    assert disc.target_x == 3.0
    assert disc.target_y == 4.0


def test_set_position(setup: Disc) -> None:
    disc = setup
    new_position = {"x": 10.0, "y": 12.0, "z": 14.0}
    disc.set_position(**new_position)
    assert disc.current_position == DiscPosition(x=10.0, y=12.0, z=14.0)


def test_move(mocker: MockerFixture, setup: Disc) -> None:
    disc = setup
    start_position = disc.current_position

    trajectory = [
        start_position,
        DiscPosition(x=2.0, y=3.0, z=4.0),
        DiscPosition(x=3.0, y=4.0, z=0.0, ax=1.0),
    ]
    disc.trajectory = trajectory
    disc.grounded = mocker.MagicMock(wraps=disc.grounded)

    # Does not move if not aired
    disc.move()
    assert disc.current_position == start_position

    # Does move when aired
    disc.release()
    disc.move()
    assert disc.current_step == 1
    assert disc.current_position == trajectory[1]

    # Does ground when z <= 0
    disc.move()
    disc.grounded.assert_called_once()
    assert disc.current_step == 0
    assert disc.state == DiscState.GROUNDED
    assert disc.trajectory == []
    assert disc.current_position == DiscPosition(x=3.0, y=4.0, z=0.0)


def test_release(setup: Disc) -> None:
    disc = setup
    disc.release()
    assert disc.state == DiscState.AIRED


def test_grounded(mocker: MockerFixture, setup: Disc) -> None:
    disc = setup
    position_mock = mocker.MagicMock(spec_set=DiscPosition)
    disc.current_position = position_mock
    disc._reset_disc_state = mocker.MagicMock()
    disc.grounded()

    position_mock.ground.assert_called_once()
    disc._reset_disc_state.assert_called_once()


def test_reset(mocker: MockerFixture, setup: Disc) -> None:
    disc = setup
    position_mock = mocker.MagicMock(spec_set=DiscPosition)
    disc.current_position = position_mock
    disc._reset_disc_state = mocker.MagicMock()
    disc.reset()

    position_mock.reset.assert_called_once()
    disc._reset_disc_state.assert_called_once()


def test__reset_disc_state(setup: Disc) -> None:
    disc = setup
    disc.release()
    disc.current_step = 10
    disc.trajectory = [DiscPosition()]
    disc._reset_disc_state()

    assert disc.state == DiscState.GROUNDED
    assert disc.current_step == 0
    assert disc.trajectory == []



import pytest
from pytest_mock.plugin import MockerFixture

from flying_discs.morrison.morrison_position import DiscMorrisonPosition


@pytest.fixture
def setup() -> DiscMorrisonPosition:
    return DiscMorrisonPosition(
        x=1.0,
        y=2.0,
        z=3.0,
        vx=4.0,
        vy=5.0,
        vz=6.0,
        ax=7.0,
        ay=8.0,
        az=9.0,
        d=10.0,
        vd=11.0,
        ad=12.0,
    )


def test_ground(setup: DiscMorrisonPosition) -> None:
    position = setup
    position.ground()
    assert position == DiscMorrisonPosition(x=1.0, y=2.0)


def test_reset(setup: DiscMorrisonPosition) -> None:
    position = setup
    position.reset()
    assert position == DiscMorrisonPosition()
